/*
 * (C) Copyright 2000-2004
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* CC1000 boardinit */

#if 0
#define DEBUG
#endif

#include <common.h>
#include <mpc8xx.h>
#ifdef CONFIG_PS2MULT
#include <ps2mult.h>
#endif

/* ------------------------------------------------------------------------- */

static long int dram_size (long int, long int *, long int);

/* ------------------------------------------------------------------------- */

#define	_NOT_USED_	0xFFFFFFFF

const uint sdram_table_a[] =
{
	/* From board */
	0xcf0ffc04, 0x0f0fec00, 0x3f0fec07, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xff0fcc04, 0x0f0fde04, 0xff0fcc00, 0xff0fcc07,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0x0fffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff
};

const uint sdram_table_b[] =
{
	0x1f0dcc04, 0xeeff8c04, 0x01ba4c04, 0x1ff5cc00,
	0xffffcc47, 0xffffffff, 0xffffffff, 0xffffffff,
	0x1f0dcc04, 0xeeff8c04, 0x10ff4c04, 0xf0ffcc00,
	0xf0ffcc00, 0xf1ffcc00, 0xffffcc00, 0xffffcc45,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0x1f0dcc04, 0xeefb8c04, 0x01a24c00, 0x1ff5cc47,
	0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
	0x1f0dcc04, 0xeefb8c00, 0x10f74c00, 0xf0ffcc00,
	0xf0ffcc00, 0xf1ffcc47, 0xffffffff, 0xffffffff,
	0x1ffd4cb4, 0xffffcc34, 0xffffcc34, 0xffffcc34,
	0xffffcc34, 0xffffcc34, 0xffffccb4, 0xffffcc35,
	0x1ffd4c04, 0xffffcc04, 0xffffcc04, 0xffffcc07,
	0xffffffff, 0xffffffff, 0x1ff5cc34, 0xeff58c34,
	0x1ff54c37, 0xffffffff, 0xffffffff, 0xffffffff,
	0x7fffcc07, 0xffffffff, 0xffffffff, 0xffffffff
};

/* ------------------------------------------------------------------------- */


/*
 * Check Board Identity:
 *
 * Test TQ ID string (TQM8xx...)
 * If present, check for "L" type (no second DRAM bank),
 * otherwise "L" type is assumed as default.
 *
 * Set board_type to 'L' for "L" type, 'M' for "M" type, 0 else.
 */

int checkboard (void)
{
	DECLARE_GLOBAL_DATA_PTR;

	char *s = getenv ("serial#");

	puts ("Board: Coactive Connector 1000\n");

	return (0);
}

/* ------------------------------------------------------------------------- */

long int initdram (int board_type)
{
	volatile immap_t *immap = (immap_t *) CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;
	long int size8, size9, size10;
	long int size_b0 = 0;
	long int size_b1 = 0;

	memctl->memc_mamr = 0x00041111;
	memctl->memc_mbmr = 0x0f823111;

	upmconfig (UPMA, (uint *) sdram_table_a,
			   sizeof (sdram_table_a) / sizeof (uint));
	upmconfig (UPMB, (uint *) sdram_table_b,
			   sizeof (sdram_table_b) / sizeof (uint));

/*
	memctl->memc_mptpr = CFG_MPTPR_1BK_4K;
*/

	/*
	 * Map controller banks 1 and 4 to the SDRAM banks 0 and 1 at
	 * preliminary addresses - these have to be modified after the
	 * SDRAM size has been determined.
	 */

	/* From reverse engineering */
	memctl->memc_or1 = 0xffc00a00;	/* Modified, was 0xff800a00 */
	memctl->memc_br1 = 0x000000c1;

	memctl->memc_mcr = 0x80802130;
	memctl->memc_mar = 0x00000088;
	memctl->memc_mcr = 0x80802136;
	memctl->memc_mcr = 0x80802828;

	memctl->memc_or4 = 0xffc00a00;
	memctl->memc_br4 = 0x004000c1;

	memctl->memc_mcr = 0x80808130;
	memctl->memc_mar = 0x00000088;
	memctl->memc_mcr = 0x80808136;
	memctl->memc_mcr = 0x80808828;

	return (0x00800000);
}

/* ------------------------------------------------------------------------- */

/*
 * Check memory range for valid RAM. A simple memory test determines
 * the actually available RAM size between addresses `base' and
 * `base + maxsize'. Some (not all) hardware errors are detected:
 * - short between address lines
 * - short between data lines
 */

static long int dram_size (long int mamr_value, long int *base, long int maxsize)
{
	volatile immap_t *immap = (immap_t *) CFG_IMMR;
	volatile memctl8xx_t *memctl = &immap->im_memctl;

	memctl->memc_mamr = mamr_value;

	return (get_ram_size(base, maxsize));
}

/* ------------------------------------------------------------------------- */

#ifdef CONFIG_PS2MULT

#ifdef CONFIG_HMI10
#define BASE_BAUD ( 1843200 / 16 )
struct serial_state rs_table[] = {
	{ BASE_BAUD, 4,  (void*)0xec140000 },
	{ BASE_BAUD, 2,  (void*)0xec150000 },
	{ BASE_BAUD, 6,  (void*)0xec160000 },
	{ BASE_BAUD, 10, (void*)0xec170000 },
};

#ifdef CONFIG_BOARD_EARLY_INIT_R
int board_early_init_r (void)
{
	ps2mult_early_init();
	return (0);
}
#endif
#endif /* CONFIG_HMI10 */

#endif /* CONFIG_PS2MULT */

/* ---------------------------------------------------------------------------- */
/* HMI10 specific stuff								*/
/* ---------------------------------------------------------------------------- */
#ifdef CONFIG_HMI10

int misc_init_r (void)
{
# ifdef CONFIG_IDE_LED
	volatile immap_t *immap = (immap_t *) CFG_IMMR;

	/* Configure PA15 as output port */
	immap->im_ioport.iop_padir |= 0x0001;
	immap->im_ioport.iop_paodr |= 0x0001;
	immap->im_ioport.iop_papar &= ~0x0001;
	immap->im_ioport.iop_padat &= ~0x0001;	/* turn it off */
# endif
	return (0);
}

# ifdef CONFIG_IDE_LED
void ide_led (uchar led, uchar status)
{
	volatile immap_t *immap = (immap_t *) CFG_IMMR;

	/* We have one led for both pcmcia slots */
	if (status) {				/* led on */
		immap->im_ioport.iop_padat |= 0x0001;
	} else {
		immap->im_ioport.iop_padat &= ~0x0001;
	}
}
# endif
#endif	/* CONFIG_HMI10 */

/* ---------------------------------------------------------------------------- */
/* NSCU specific stuff								*/
/* ---------------------------------------------------------------------------- */
#ifdef CONFIG_NSCU

int misc_init_r (void)
{
	volatile immap_t *immr = (immap_t *) CFG_IMMR;

	/* wake up ethernet module */
	immr->im_ioport.iop_pcpar &= ~0x0004; /* GPIO pin	*/
	immr->im_ioport.iop_pcdir |=  0x0004; /* output		*/
	immr->im_ioport.iop_pcso  &= ~0x0004; /* for clarity	*/
	immr->im_ioport.iop_pcdat |=  0x0004; /* enable		*/

	return (0);
}
#endif	/* CONFIG_NSCU */

/* ------------------------------------------------------------------------- */
